// -------------------------------------------------------------
// 3d code by Georgy Moshkin (Delphi)
// licence: public domain
// -------------------------------------------------------------
unit mydepthcollision;

interface

uses OpenGL;

const ZBuf_w = 128;
      ZBuf_h = 128;

const ZAreaElementDimension_x = 5;
      ZAreaElementDimension_z = 5;

const ZAreaDimension_x = 15;
      ZAreaDimension_z = 15;

const ZLayer_w = ZBuf_w * (ZAreaDimension_x div ZAreaElementDimension_x); //1024;
      ZLayer_h = ZBuf_h * (ZAreaDimension_z div ZAreaElementDimension_z); //1024;

const ZArea_x = ZLayer_w div ZBuf_w;
      ZArea_y = 6;
      ZArea_z = ZLayer_h div ZBuf_h;

const my_NEAR=1;
      my_FAR=1000;

const EPSILON=0.05;      

type TZbuffer=packed array[0..ZBuf_w-1,0..ZBuf_h-1] of glFloat;

type TZLayerElement=record
                     zDATA        : TZbuffer;
                     captureHeight: single;
                     parent_i     : integer;
                     parent_j     : integer;
                     parent_k     : integer;
                    end;

type TZLayerElements=array of TZLayerElement;

type TLinkToZLayerElement=record
                           child_index   : integer;
                           captureHeight : single;
                          end;

type TZArea=array[0..ZArea_x-1,
                  0..ZArea_y-1,
                  0..ZArea_z-1] of TLinkToZLayerElement;


var ZBuffer:TZbuffer;
    ZArea:TZArea;
    ZLayerElements:TZLayerElements;

var min_Y, max_Y:single;

var depthcollisionFirstRun:boolean=true;

var z_i, z_j, z_k:integer;


procedure BeforeRender;
function AfterRender:boolean;
procedure SetOrthoMode;
procedure SetOrthoCamera(Area_i,Area_j,Area_k:integer);
procedure GetCurrentZBuffer;
procedure StoreBufferData(Area_i,Area_j,Area_k:integer);
procedure SaveData(fname:string);
procedure ShowCollisionDataOpengl;
function isequal(zbuf1,zbuf2:TZbuffer):boolean;
function getDIFF(zbuf1,zbuf2:TZbuffer):single;

implementation

procedure BeforeRender;
begin

 if depthcollisionFirstRun=true then
  begin
   depthcollisionFirstRun:=false;
   z_i:=0;
   z_j:=0;
   z_k:=0;
   SetOrthoMode;
  end;

 glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
 glLoadIdentity();

 SetOrthoCamera(z_i,z_j,z_k);

end;

function AfterRender:boolean;
var AllDone:boolean;
begin
 AllDone:=false;

 GetCurrentZBuffer;
 StoreBufferData(z_i,z_j,z_k);

 inc(z_i);
 if z_i>ZArea_x-1 then
  begin
   z_i:=0;
   inc(z_j);
   if z_j>ZArea_y-1 then
    begin
     z_j:=0;
     inc(z_k);
     if z_k>ZArea_z-1 then
      begin
       AllDone:=true;
       depthcollisionFirstRun:=true;
      end;
    end;
  end;

  result:=AllDone;
end;

procedure SetOrthoMode;
begin
 glMatrixMode(GL_PROJECTION);
 glPushMatrix();
 glLoadIdentity();

 glOrtho(-ZAreaElementDimension_x/2, ZAreaElementDimension_x/2,
         -ZAreaElementDimension_z/2, ZAreaElementDimension_z/2,
             my_NEAR, my_FAR);

 glMatrixMode(GL_MODELVIEW);
 glPushMatrix();
 glLoadIdentity();

 glPopMatrix;
 glPopMatrix;

 glViewport(0, 0, zbuf_W,zbuf_H); 
end;

procedure SetOrthoCamera(Area_i,Area_j,Area_k:integer);
var tmpX,tmpY,tmpZ:single;
begin

  tmpY:=min_Y + ( Area_j/(ZArea_y-1) ) * (max_y - min_y);

  tmpX:=(Area_i - (ZArea_x-1)/2)*ZAreaElementDimension_x;
  tmpZ:=(Area_k - (ZArea_z-1)/2)*ZAreaElementDimension_z;

//  tmpx:=0;
//  tmpy:=5;
//  tmpz:=0;

  gluLookAt(tmpX, tmpY  , tmpZ,
            tmpX, tmpY-1, tmpZ,
            1,    0,      0);
end;

procedure GetCurrentZBuffer;
begin
  glReadPixels(0, 0, zbuf_W, zbuf_H, GL_DEPTH_COMPONENT, GL_FLOAT, @Zbuffer);
end;

procedure StoreBufferData(Area_i,Area_j,Area_k:integer);
var k:integer;
    tmpY:single;

    i,j:integer;
    FOUNDEQUAL:boolean;
    EQUALNUMBER:integer;
    EQUALDIFF:single;
    ANOTHERHEIGHT:single;
begin

 FOUNDEQUAL:=FALSE;

 for i:=0 to length(ZLayerElements)-1 do
  if isequal(ZBuffer,ZLayerElements[i].zDATA) then
   begin
    FOUNDEQUAL:=true;
    EQUALNUMBER:=i;
    EQUALDIFF:=getDIFF(ZBuffer,ZLayerElements[i].zDATA);
    ANOTHERHEIGHT:=ZLayerElements[i].captureHeight;
   end;
   
 if not FOUNDEQUAL then
  begin
   k:=length(ZLayerElements);
   setlength(ZLayerElements,k+1);

   tmpY:=min_Y + ( Area_j/(ZArea_y-1) ) * (max_y - min_y);

   ZLayerElements[k].zDATA:=ZBuffer;
   ZLayerElements[k].captureHeight:=tmpY;
   ZLayerElements[k].parent_i:=Area_i;
   ZLayerElements[k].parent_j:=Area_j;
   ZLayerElements[k].parent_k:=Area_k;

   ZArea[Area_i,Area_j,Area_k].child_index:=k;
   ZArea[Area_i,Area_j,Area_k].captureHeight:=tmpY;
   // ZArea[Area_i,Area_j,Area_k].optionalConst:=-1;
 end
  else
 begin
  ZArea[Area_i,Area_j,Area_k].child_index:=EQUALNUMBER;
  ZArea[Area_i,Area_j,Area_k].captureHeight:=ANOTHERHEIGHT;
 end;

end;

procedure SaveData(fname:string);
var f:file;
var i,j,k:integer;
begin
 assignFile(f,fname);
 rewrite(f,1);


 for i:=0 to ZArea_x-1 do
  for j:=0 to ZArea_y-1 do
   for k:=0 to ZArea_z-1 do
    blockWrite(f,ZArea[i,j,k],SizeOf(TLinkToZLayerElement));

//  blockWrite(f,ZArea,SizeOf(TLinkToZLayerElement)*ZArea_x*ZArea_y*ZArea_Z);

   for i:=0 to length(ZLayerElements)-1 do
   begin
    blockWrite(f,ZLayerElements[i],sizeof(TZLayerElement))
   end;


 closeFile(f);
end;

procedure ShowCollisionDataOpengl;
var Area_i,Area_j,Area_k:integer;
var tmpX,tmpY,tmpZ:single;
var u,v:integer;
begin



 for Area_i:=0 to ZArea_x-1 do
  for Area_j:=0 to ZArea_y-1 do
   for Area_k:=0 to ZArea_z-1 do
    begin
//     tmpY:=min_Y + ( Area_j/(ZArea_y-1) ) * (max_y - min_y);
     tmpY:=ZArea[Area_i,Area_j,Area_k].captureHeight;

     tmpX:=(Area_i - (ZArea_x-1)/2)*ZAreaElementDimension_x;
     tmpZ:=(Area_k - (ZArea_z-1)/2)*ZAreaElementDimension_z;

     glColor3f(1,1,1);
     glBegin(GL_LINES);
      glVertex3f(tmpX,tmpY,tmpZ);
      glVertex3f(tmpX,tmpY-0.1,tmpZ);
     glEnd;

     glPointSize(3);
     glColor3f(1,0,0);
     glBegin(GL_POINTS);
      glVertex3f(tmpX,tmpY,tmpZ);
     glEnd;

     glColor3f(Area_j/ZArea_y,
               Area_j/ZArea_y,
               Area_j/ZArea_y);

     glPointSize(1);
     glBegin(GL_POINTS);
     for u:=0 to ZBuf_w-1 do
      for v:=0 to ZBuf_h-1 do
       begin

        glVertex3f(tmpx-ZAreaElementDimension_x/2+u*ZAreaElementDimension_x/ZBuf_w,
        tmpy - ZLayerElements[ZArea[Area_i,Area_j,Area_k].child_index].zDATA[u,v]*(MY_FAR-MY_NEAR)-1,
        tmpz-ZAreaElementDimension_z/2+v*ZAreaElementDimension_z/ZBuf_h);

       end;
       glEnd;





    end;
end;

function isequal(zbuf1,zbuf2:TZbuffer):boolean;
var k,l:integer;
    diff:single;
    ISequal:boolean;
begin
 diff:=zbuf2[0,0]-zbuf1[0,0];
 ISequal:=true;

 for k:=0 to ZBuf_w-1 do
  for l:=0 to ZBuf_h-1 do
   begin
    if abs(zbuf2[k,l]-zbuf1[k,l])>EPSILON then ISequal:=false;
   end;

 result:=ISequal;   
end;

function getDIFF(zbuf1,zbuf2:TZbuffer):single;
begin
 result:=zbuf2[0,0]-zbuf1[0,0];
end;

end.
